#include "StdAfx.h"
#include "Persist.h"
#include "Utility.h"

void Persist::PrePersist(const CString& name) {
	if (IsLoading()) {
		//make sure the correct field exists
		if (FindSubstring(mData, name + '=', ffMatchCase, mOffset) != mOffset) {
			CString msg;
			msg.Format(TEXT("Could not find field '%s' when loading file '%s'. The file may be corrupted. You should either replace the file with a valid copy or delete the file from disk."), name, mFileName);
			throw Exception(ErrPersistedDataNotFound, msg);
		}

		mOffset += name.GetLength() + 1;//+1 for =
	} else {
		mData += name;
		mData += '=';
	}
}

Persist::Persist(const CString& fileName, bool loading) : mFileName(fileName), mLoading(loading), mOffset(0), mVersion(CurPersistVer) {
	//can't load if the file doesn't exist
	ASSERT(!IsLoading() || FileExists(fileName));

	if (IsLoading())
		mData = FileToString(mFileName);

	PersistNumber(TEXT("Version"), mVersion);

	if (loading && mVersion > CurPersistVer)
		throw Exception(ErrNewerFileVersion);

	PersistNewLine();
}

Persist::~Persist() {
	if (!IsLoading()) {
		//convert to CRLF so it can be read by normal windows text editors
		CString fileData = ConvertLFtoCRLF(mData);
		StringToFile(mFileName, fileData);
	}
}

//puts a new line in persistence file to make it more readable
void Persist::PersistNewLine() {
	if (IsLoading()) {
		long end = mData.Find('\n', mOffset);
		if (end < 0)
			mOffset = mData.GetLength();
		else
			mOffset = end + 1;
	} else
		mData += '\n';
}

void Persist::PersistBool(const CString& name, bool& val) {
	CString stringVal = val ? TEXT("True") : TEXT("False");
	PersistString(name, stringVal);
	val = StrCmpI(stringVal, TEXT("False")) != 0;
}

void Persist::PersistNumber(const CString& name, int& val) {
	PrePersist(name);

	if (IsLoading()) {
		sscanf(GetReadPtr(), TEXT("%d"), &val);
	} else {
		CString temp;
		temp.Format(TEXT("%d"), val);
		WriteData(temp);
	}

	PostPersist();
}

void Persist::PersistString(const CString& name, CString& val) {
	PrePersist(name);

	if (IsLoading()) {
		val = EmptyStr;
		long offset = 0;
		while (true) {
			char ch = NULL;
			sscanf(GetReadPtr() + offset, TEXT("%c"), &ch);
			if (ch == NULL || ch == '\n')
				break;

			val += ch;
			++offset;
		}
	} else {
		//Can't persist a string with new lines in it. That will
		//need to be persisted a several strings.
		ASSERT(val.Find('\n') < 0);
		WriteData(val);
	}

	PostPersist();
}

void Persist::PersistVal(const CString& name, LOGFONT& lf) {
	PersistNumber(name + TEXT("Height"), lf.lfHeight);
	PersistNumber(name + TEXT("Width"), lf.lfWidth);
	PersistNumber(name + TEXT("Escapement"), lf.lfEscapement);
	PersistNumber(name + TEXT("Orientation"), lf.lfOrientation);
	PersistNumber(name + TEXT("Weight"), lf.lfWeight);
	PersistBool(name + TEXT("Italic"), lf.lfItalic);
	PersistBool(name + TEXT("Underline"), lf.lfUnderline);
	PersistBool(name + TEXT("StrikeOut"), lf.lfStrikeOut);
	PersistNumber(name + TEXT("CharSet"), lf.lfCharSet);
	PersistNumber(name + TEXT("OutPrecision"), lf.lfOutPrecision);
	PersistNumber(name + TEXT("ClipPrecision"), lf.lfClipPrecision);
	PersistNumber(name + TEXT("Quality"), lf.lfQuality);
	PersistNumber(name + TEXT("PitchAndFamily"), lf.lfPitchAndFamily);
	CString faceName = lf.lfFaceName;
	PersistString(name + TEXT("FaceName"), faceName);
	StrCpy(lf.lfFaceName, faceName);
}

void Persist::PersistVal(const CString& name, HFONT& hFont) {
	LOGFONT lf;
	MemClear(lf);

	if (IsLoading()) {
		PersistVal(name, lf);
		hFont = CreateFontIndirect(&lf);
	} else {
		//convert hFont to LOGFONT
		VERIFY(::GetObject(hFont, sizeof(lf), &lf) == sizeof(lf));
		PersistVal(name, lf);
	}
}

void Persist::PersistVal(const CString& name, RECT& rc) {
	PersistNumber(name + TEXT("Top"), rc.top);
	PersistNumber(name + TEXT("Left"), rc.left);
	PersistNumber(name + TEXT("Bottom"), rc.bottom);
	PersistNumber(name + TEXT("Right"), rc.right);
}
