#ifndef ConfigDataH
#define ConfigDataH

#include "EditCtrl.h"

typedef vector<LanguagePtr> Languages;

class LanguageMgr {
private:
public:
	Languages mLangs;

	LanguagePtr FindLanguageByExtension(const CString& ext, LanguagePtr skipLang = NULL) const;
	bool LanguageExists(const CString& name) const;
	void Load();
	void Save();
	CString BuildFilter() const;
	CString FilterIndexToFilterName(long index) const;
	long FilterNameToFilterIndex(const CString& name) const;
};

class RecentFileList {
private:
	long mMax;
	StringVector mFiles;

	void ValidateSize();
	void PergeNonExistantFiles();

public:
	RecentFileList() : mMax(15) {}

	long GetMax() const {return mMax;}
	void SetMax(long val);

	void Add(const CString& fileName);
	void Remove(const CString& fileName);
	const StringVector& GetItems() const {return mFiles;}
	HMENU CreateMenu(CommandIdCreator& idCreator) const;
	void DoPersist(Persist& p);
};

static LPCTSTR FilePathMacro = TEXT("%FilePath%");
static LPCTSTR FileDirMacro = TEXT("%FileDir%");
static LPCTSTR FileNameMacro = TEXT("%FileName%");
static LPCTSTR FileExtMacro = TEXT("%FileExt%");
static LPCTSTR CurDirMacro = TEXT("%CurDir%");
static CString PromptMacroStart = TEXT("%Prompt:");

CString GetCommandString(const CString& curFileName, const CString& format, bool displayStringOnly = false);

class Command {
public:
	CString mMenuItemName;
	CString mCommand;
	CString mArguments;
	CString mInitDir;
	bool mSaveFile;
	StringVector mExts;
	Accel mAccel;
	bool mIsFilter;

	Command(const CString& command = EmptyStr, const CString& menuItemName = TEXT("New &Tool"), const CString& arguments = Quote(FilePathMacro), const CString& initDir = FileDirMacro, bool saveFile = true, VirtualKey key = 0, KeyModifiers mods = kmNone) :
		mMenuItemName(menuItemName), mCommand(command), mArguments(arguments), mInitDir(initDir),
			mSaveFile(saveFile), mAccel(key, mods, 0), mIsFilter(false) {}

	bool HasExtension(const CString& ext) const {return mExts.HasString(ext, false);}
	void DoPersist(Persist& p);
};

class Commands : public Vector<Command> {
public:
	AccelTable mAccelTable;

	void AddCommandsToMenu(HMENU hMenu, CommandIdCreator& idCreator);
	Command* FindCommand(UINT cmd);
	void DoPersist(Persist& p);
};

//Subset of FindReplaceData that's used for searching.
//It's useful to keep "search criteria" data separate from
//FindReplaceData so you can store the search criteria
//and know when it changed.
struct SearchData {
	CString findStr;
	FindFlags flags;
	bool selectionOnly;
	bool escapedChars;
	bool newSearch;

	SearchData() :
		flags(ffIgnoreCase), selectionOnly(false), escapedChars(true), newSearch(true) {}

	bool Changed(const SearchData& rhs, bool &changedDirection) {
		if (newSearch) {
			changedDirection = false;
			return true;
		} else {
			changedDirection = (flags & ffReverse) != (rhs.flags & ffReverse);
			bool changedNonDirFlag = (flags & ~ffReverse) != (rhs.flags & ~ffReverse);

			return
				changedNonDirFlag ||
				selectionOnly != rhs.selectionOnly ||
				escapedChars != rhs.escapedChars ||
				StrCmpEx(findStr, rhs.findStr, (flags & ffMatchCase) != 0) != 0;
		}
	}
};

struct FindReplaceData : SearchData {
	//the following are never read by the dialog and so don't need to be initialized
	bool findNext;
	bool replaceCurrent;
	bool replaceAll;
	CString replaceStr;

	StringVector findStrings;
	StringVector replaceStrings;
	bool closeOnFind;//persisted

	FindReplaceData() :
		closeOnFind(true) {}

	void DoPersist(Persist& p);
};

enum SelColorKinds {scInvert, scSystem};

class ColorScheme : public TokenColors {
private:
	CString mName;
	SelColorKinds mSelColor;

public:
	ColorScheme(const CString& name = EmptyStr, SelColorKinds sck = scInvert) : mName(name), mSelColor(sck) {}

	const CString& GetName() const {return mName;}
	void SetName(const CString& name) {mName = name;}
	SelColorKinds GetSelColor() const {return mSelColor;}
	void SetSelColor(SelColorKinds scKind) {mSelColor = scKind;}

	virtual void DoPersist(Persist& p);
};
PTR_DEF(ColorScheme);

class ColorSchemes : public Vector<ColorSchemePtr> {
private:
	long mCurScheme;

public:
	ColorSchemes() : mCurScheme(0) {}
	ColorSchemes(const ColorSchemes& rhs) {*this = rhs;}

	ColorSchemes& operator=(const ColorSchemes& rhs);

	ColorSchemePtr GetCurColorScheme() const {
		return mCurScheme >= 0 && !empty() ? Item(mCurScheme) : NULL;
	}

	long GetCurScheme() const {return mCurScheme;}

	void SetCurScheme(long val) {
		ATLASSERT(-1 <= val && val < (long)size());
		mCurScheme = val;
	}

	void DoPersist(Persist& p);	
};

static LPCTSTR DateMacro = TEXT("%Date%");
static LPCTSTR LongDateMacro = TEXT("%LongDate%");
static LPCTSTR TimeMacro = TEXT("%Time%");
static LPCTSTR LongTimeMacro = TEXT("%LongTime%");

CString GetTimeDateString(const CString& format);

static const int MAX_CUSTOM_COLORS = 16;

static const CString AllFilesFilter = TEXT("All Files|*.*|");

struct ConfigData {
	CString mDefExt;
	CString mTimeDateFormat;
	bool mRemoveTrailingWhitespaceOnSave;
	NewLineTypes mNewLineType;
	bool mAutoIndent;
	long mTabWidth;
	RecentFileList mRecentFileList;
	Commands mCommands;
	FindReplaceData mFindReplaceData;
	CRect mWinPos;
	bool mMaximized;
	bool mShowLineNumbers;
	bool mTabInsertsSpaces;
	bool mShowWhitespace;
	bool mWordWrap;

	//I persist a name for the open file filter instead of its
	//index because the index would be wrong if languages were
	//added, remove, or reordered
	CString mOpenFilterName;

	CFont mFont;
	ColorSchemes mColorSchemes;
	LanguageMgr mLangMgr;
	COLORREF mCustomColors[MAX_CUSTOM_COLORS];
	AccelTable mAccelTable;

	//this isn't persisted, but needs to be shared
	CString mTempClip;

	void InitDefaultValues();
	void DoPersist(bool load);
	void Load() {DoPersist(true);}
	void Save() {DoPersist(false);}
	void SetEdit(EditCtrl& edit);
	bool CanSwitchClipboard() const;
	void SwitchClipboard(HWND hwnd);

	bool Execute(const Command& cmd, const CString& fileName, HWND hwnd, CString& filterText = CString(EmptyStr)) const;
	void ExecuteHelp(const CString& topic, HWND hwnd) const;

	long GetOpenFilterIndex() const {return mLangMgr.FilterNameToFilterIndex(mOpenFilterName);}
	void SetOpenFilterIndex(long index) {mOpenFilterName = mLangMgr.FilterIndexToFilterName(index);}

private:
	void AddConfigDataForOldVersions(long version);
};

extern ConfigData gConfigData;

//The following isn't actually config data. It just needs to be shared
//across dialogs and isn't general enough to be in Common.

static const CString PROGRAM_NAME = TEXT("Lite Edit");
static const CString PROGRAM_LONGNAME = PROGRAM_NAME + TEXT(" 2.0.1 [Red]");

#endif